# -*- Mode: Python; encoding: utf-8 -*-

'''
A widget showing a MathML expression
'''

if __name__ == '__main__':
    import sys
    sys.path.insert(0, "../../lib/pymathml")

import gtk, gtk.gdk
import cairo
import pangocairo
import math
import gconf

import mathml.xml.sax_parser
from mathml.backend.pangocairo_mathml import *

__all__ = ['MathMLView']

def _get_resolution():
    resolution_mm = (gtk.gdk.screen_height() /
                     float(gtk.gdk.screen_height_mm()))
    resolution_pt = resolution_mm * 0.3514598
    return resolution_pt


class MathMLView(gtk.Widget):
    __gtype_name__ = "MathMLView"

    def __init__(self):
        gtk.Widget.__init__(self)
        self.rendered_mathml = None # cairo surface containing pre-rendered mathml
        self.mathml_tree = None
        self.axis = None
        self._sax_handler = None

    def do_realize(self):
        self.set_flags(self.flags() | gtk.REALIZED)
        self.window = gtk.gdk.Window(
            self.get_parent_window(),
            width=self.allocation.width,
            height=self.allocation.height,
            window_type=gtk.gdk.WINDOW_CHILD,
            wclass=gtk.gdk.INPUT_OUTPUT,
            event_mask=self.get_events() | gtk.gdk.EXPOSURE_MASK)
        self.window.set_user_data(self)
        self.style.attach(self.window)
        self.style.set_background(self.window, gtk.STATE_NORMAL)
        self.window.move_resize(*self.allocation)

    def do_unrealize(self):
        self.window.set_user_data(None)

    def do_size_request(self, requisition):
        if self.rendered_mathml is None:
            requisition.width = 0
            requisition.height = 0
        else:
            requisition.width = self.rendered_mathml.get_width()
            requisition.height = self.rendered_mathml.get_height()

    def do_size_allocate(self, allocation):
        self.allocation = allocation
        if self.flags() & gtk.REALIZED:
            self.window.move_resize(*allocation)

    def do_expose_event(self, event):
        if self.rendered_mathml is None:
            return
        cr = self.window.cairo_create()
        cr.set_source_surface(self.rendered_mathml, 0, 0)
        cr.rectangle(event.area)
        cr.clip()
        cr.paint()

    def parse_string(self, mathml_xml):
        ## we create a temporary context just to get the dimmensions
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, 0, 0)
        cr = pangocairo.CairoContext(cairo.Context(surface))
        pl = PangoCairoPlotter(cr, _get_resolution())
        self.mathml_tree = mathml.xml.sax_parser.parseString(mathml_xml, pl)
        self.render()

    def do_style_set(self, style):
        if style is not None:
            gtk.Widget.do_style_set(self, style)
            self.render()

    def render(self):
        if self.mathml_tree is None:
            self.rendered_mathml = None
            return
        self.ensure_style()
        margin = 4
        root = self.mathml_tree
        root.x0 = 0
        root.y0 = 0
        root.setAttribute("fontsize", "18pt")
        root.update()
        ## now create the real surface, with the correct dimmensions
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                     int(math.ceil(root.width)) + margin*2,
                                     int(math.ceil(root.height)) + margin*2)
        cr = gtk.gdk.CairoContext(cairo.Context(surface))
        cr.set_source_color(self.style.base[gtk.STATE_NORMAL])
        cr.paint()
        cr.set_source_color(self.style.fg[gtk.STATE_NORMAL])
        root.plotter.set_cairo_context(pangocairo.CairoContext(cr))
        root.x0 = margin
        root.y0 = margin - surface.get_height()
        root.render()
        root.plotter.close()
        self.axis = surface.get_height() - (margin + root.axis)
        ## save the image surface for usage in expose events
        self.rendered_mathml = surface
        
        self.queue_resize()
        self.queue_draw()

    def begin_sax_parse(self):
        ## we create a temporary context just to get the dimmensions
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, 0, 0)
        cr = pangocairo.CairoContext(cairo.Context(surface))
        pl = PangoCairoPlotter(cr, _get_resolution())
        self.render()
        self._sax_handler = mathml.xml.sax_parser.MathMLHandler(pl)
        return self._sax_handler

    def end_sax_parse(self):
        root = self._sax_handler.buildTree()
        self._sax_handler = None
        font = pango.FontDescription(
            gconf.client_get_default().get_string("/desktop/gnome/interface/font_name"))
        root.setAttribute("fontsize", "%ipt" % int(font.get_size()*1.5/pango.SCALE))
        root.setAttribute("minsize", "%ipt" % (0.707*font.get_size()//pango.SCALE))
        self.mathml_tree = root
        self.render()
        

gtk.rc_parse_string('''
style "mathmlview" = "GtkTextView" {
#  bg[NORMAL] = "red"
}
class "MathMLView" style "mathmlview"
''')


def test():
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    view = MathMLView()
    window.add(view)
    view.show()
    view.parse_string("""
    <math>
    <mrow>
    <mfrac>
      <mn>1</mn>
      <mrow><mi>&#x03c3;</mi> <mo>&#8290;</mo>
        <msqrt><mn>2</mn><mo>&#8290;</mo><mi>&#x03c0;</mi></msqrt>
      </mrow>
    </mfrac>
    <mo>&#x2062;</mo>
    <msup>
    <mi>e</mi>
      <mrow>
      <mo>-</mo>
        <msup>
          <mrow><mo>(</mo>
            <mfrac>
              <mrow><mi>x</mi><mo>-</mo><mi>&#x00b5;</mi></mrow>
              <mrow><msqrt><mn>2</mn></msqrt><mo>&#8290;</mo><mi>&#x03c3;</mi></mrow>
            </mfrac>
          <mo>)</mo></mrow>
          <mn>2</mn>
        </msup>
      </mrow>
    </msup>
    </mrow>
    </math>
""")
    #window.set_property("border-width", 4)
    window.connect("destroy", lambda unused: gtk.main_quit())
    window.show()
    gtk.main()

if __name__ == '__main__':
    test()
