import misc
import gobject
import pango
import gtk
import xml.sax, xml.sax.handler
import re
import warnings
from cStringIO import StringIO
from mathmlview import MathMLView
import gnomevfs
import gconf

#-- 
__all__ = ['Display']


whitespace_rx = re.compile("\\s+")
allwhitespace_rx = re.compile("^\\s*$")



class HelpHandler(xml.sax.handler.ContentHandler):
    def __init__(self, textview, startiter):
        assert isinstance(textview, gtk.TextView)

        ## we ignore everything up to the first <div> inside the <html>
        self.ignore_content = True
        self.level = 0
        
        self.textbuf   = textview.get_buffer()
        self.textview  = textview
        self.iter      = startiter

        self.italic    = 0
        self.bold      = 0
        self.sup       = 0
        self.sub       = 0
        self.underline = 0
        self.tt        = 0
        self.paragraph = 0

        self.text      = ''
        self.indent_level = 0
        self.mathml    = 0
        self.klass     = []

    def get_style_tags(self):
        tags = []
        if self.italic: tags.append("italic")
        if self.bold: tags.append("bold")
        if self.sup: tags.append("sup")
        if self.sub: tags.append("sub")
        if self.tt: tags.append("tt")
        if self.paragraph: tags.append("paragraph")
        try:
            klass = self.klass[-1]
        except IndexError:
            klass = None
        if klass is None: pass
        elif klass == 'input': tags.append("question")
        elif klass == 'output': tags.append("answer")
        elif klass == 'plot': tags.append("justify-center")
        else: warnings.warn("Unrecognized div class '%s'" % klass)
        return tags

    def flush_text(self):
        if not self.text: return
        tags = self.get_style_tags()
        self.textbuf.insert_with_tags_by_name(self.iter, self.text,
                			      "help-body", *tags)
        self.text = ''
        
    def characters(self, content):
        if self.ignore_content:
            return
        if self.mathml:
            return self.mathml_handler.characters(content)
        if allwhitespace_rx.match(content) is not None: return
        if self.text: self.text += ' '
        self.text += whitespace_rx.sub(' ', content)

    def startElement(self, name, attrs):
        self.level += 1
        if self.level == 3 and name == 'div':
            self.ignore_content = False
        if self.ignore_content:
            return
        
        if self.mathml:
            self.mathml += 1
            return self.mathml_handler.startElement(name, attrs)

        try:
            self.klass.append(attrs['class'])
        except KeyError:
            try:
                self.klass.append(self.klass[-1])
            except IndexError:
                self.klass.append(None)

        self.flush_text()
        if   name == 'i':
            self.italic += 1
        elif name == 'b':
            self.bold += 1
        elif name == 'sup':
            self.sup += 1
        elif name == 'sub':
            self.sub += 1
        elif name == 'tt':
            self.tt += 1
        elif name == 'br':
            pass # handled in endElement
        elif name == 'p':
            if not self.iter.starts_line():
                self.textbuf.insert(self.iter, "\n")
            self.paragraph += 1
        elif name == 'div':
            if not self.iter.starts_line():
                self.textbuf.insert(self.iter, "\n")
        elif name == 'ul':
            self.indent_level += 1
        elif name == 'li':
            self.text = ' '*self.indent_level*4 + unichr(0x2022) + ' '
        elif name == 'img':
            mem = gnomevfs.read_entire_file(attrs['src'])
            loader = gtk.gdk.PixbufLoader()
            loader.write(mem); loader.close()
            self.textbuf.insert(self.iter, "\n")
            tmpmark = self.textbuf.create_mark(None, self.iter, True)
            pixbuf = loader.get_pixbuf()
            self.textbuf.insert_pixbuf(self.iter, pixbuf)
            start = self.textbuf.get_iter_at_mark(tmpmark)
            for tag in self.get_style_tags():
                self.textbuf.apply_tag_by_name(tag, start, self.iter)
            self.textbuf.delete_mark(tmpmark)
            self.textbuf.insert(self.iter, "\n")
            
        elif name == 'math':
            self.mathml_view = MathMLView()
            self.mathml_handler = self.mathml_view.begin_sax_parse()
            self.mathml_handler.startElement(name, attrs)
            self.mathml = 1
        else:
            warnings.warn("Unhandled element '%s'" % name)


    def endElement(self, name):
        if self.level == 3 and name == 'div':
            self.ignore_content = True
        self.level -= 1
        if self.ignore_content:
            return
        if self.mathml:
            self.mathml_handler.endElement(name)
            self.mathml -= 1
            if not self.mathml:
                self.mathml_view.end_sax_parse()
                self.insert_mathml()
            return

        self.flush_text()
        if   name == 'i':
            self.italic -= 1
        elif name == 'b':
            self.bold -= 1
        elif name == 'sup':
            self.sup -= 1
        elif name == 'sub':
            self.sub -= 1
        elif name == 'tt':
            self.tt -= 1
        elif name == 'p':
            if not self.iter.starts_line():
                self.textbuf.insert(self.iter, "\n")
            self.paragraph -= 1
        elif name == 'div':
            if not self.iter.starts_line():
                self.textbuf.insert(self.iter, "\n")
        elif name == 'br':
            self.textbuf.insert(self.iter, "\n")
        elif name == 'ul':
            self.indent_level -= 1
        elif name == 'li':
            self.text = '\n'
        elif name == 'img':
            pass
        else:
            warnings.warn("Unhandled element '%s'" % name)

        self.klass.pop()

    def insert_mathml(self):
        root = self.mathml_view.mathml_tree
        display = root.getAttribute('display', default='inline').str
        if   display == 'inline': block = False
        elif display == 'block': block = True
        else: raise ValueError("Invalid mathml display mode")

        if block:
            self.textbuf.insert(self.iter, "\n")
        tmpmark = self.textbuf.create_mark(None, self.iter, True)
        anchor = self.textbuf.create_child_anchor(self.iter)
        self.mathml_view.show()
        self.textview.add_child_at_anchor(self.mathml_view, anchor)
        start = self.textbuf.get_iter_at_mark(tmpmark)
        if block:
            self.textbuf.apply_tag_by_name("justify-center", start, self.iter)
        else: # inline
            tag = gobject.new(gtk.TextTag, rise=(-self.mathml_view.axis + 2)*1000)
            self.textbuf.get_tag_table().add(tag)
            self.textbuf.apply_tag(tag, start, self.iter)
        self.textbuf.delete_mark(tmpmark)
        if block:
            self.textbuf.insert(self.iter, "\n")
        self.mathml_view = None

class Display(misc.ConsoleDisplay):
    
    def __init__(self):
        self.textview = gtk.TextView()
        self.textview.show()
        self.textview.set_wrap_mode(gtk.WRAP_CHAR)
        self.textview.set_editable(False)
        buffer = self.textview.get_buffer()

        sw = gtk.ScrolledWindow()
        sw.set_property("hscrollbar-policy", gtk.POLICY_AUTOMATIC)
        sw.set_property("vscrollbar-policy", gtk.POLICY_AUTOMATIC)
        sw.set_property("border-width", 0)
        sw.add(self.textview)
        sw.show()
        self.vadjustment = sw.get_vadjustment()

        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_IN)
        frame.show()
        frame.add(sw)
        self.frame = frame
        self.__create_tags()
        self.__eob_mark = buffer.create_mark("end-of-buffer",
                			     buffer.get_end_iter(), False)

    def get_widget(self):
        return self.frame

    def __create_tags(self):
        table = self.textview.get_buffer().get_tag_table()
        # <i>
        table.add(gobject.new(gtk.TextTag, name="italic",
                	      style=pango.STYLE_ITALIC))
        # <b>
        table.add(gobject.new(gtk.TextTag, name="bold",
                	      weight=pango.WEIGHT_BOLD))
        # <sup>
        table.add(gobject.new(gtk.TextTag, name="sup",
                	      scale=0.8, rise=6000))
        # <sub>
        table.add(gobject.new(gtk.TextTag, name="sub",
                	      scale=0.8, rise=-6000))
        # <tt>
        table.add(gobject.new(gtk.TextTag, name="tt",
                	      font="Monospace"))
        # help title
        table.add(gobject.new(gtk.TextTag, name="help-title",
                	      size_points=18.0,
                	      underline=pango.UNDERLINE_SINGLE,
                	      font="Monospace"))
        # help body
        table.add(gobject.new(gtk.TextTag, name="help-body",
                	      wrap_mode=gtk.WRAP_WORD,
                	      left_margin=16, right_margin=16))
        # justify-left
        table.add(gobject.new(gtk.TextTag, name="justify-left",
                	      justification=gtk.JUSTIFY_LEFT))
        # justify-center
        table.add(gobject.new(gtk.TextTag, name="justify-center",
                	      justification=gtk.JUSTIFY_CENTER))
        # justify-right
        table.add(gobject.new(gtk.TextTag, name="justify-right",
                	      justification=gtk.JUSTIFY_RIGHT))
        # question
        table.add(gobject.new(gtk.TextTag, name="question",
                	      background="#e0e0ff",
                	      font="Monospace"))
        # answer
        table.add(gobject.new(gtk.TextTag, name="answer",
                	      background="#ffffe0",
                	      font="Monospace",
                	      left_margin=16))
        # message
        table.add(gobject.new(gtk.TextTag, name="message",
                	      #foreground="#888800",
                	      font="Monospace"))
        # paragraph
        table.add(gobject.new(gtk.TextTag, name="paragraph",
                	      pixels_above_lines=4,
                	      pixels_below_lines=4))

    def scroll_to_bottom(self):
        self.textview.scroll_to_mark(self.__eob_mark, 0, True, 0.0, 1.0)

    def display_qa(self, question, answer):
        buffer = self.textview.get_buffer()
        eob = buffer.get_end_iter()
        buffer.insert_with_tags_by_name(eob, question + "\n", "question")
        buffer.insert_with_tags_by_name(eob, answer + "\n", "answer")
        self.scroll_to_bottom()
        
    def display_help(self, title, body):
        buffer = self.textview.get_buffer()
        eob = buffer.get_end_iter()
        buffer.insert(eob, "\n")
        buffer.insert_with_tags_by_name(eob, title + "\n", "help-title")
        buffer.insert(eob, "\n")

        # xml parse
        parser = xml.sax.make_parser(['drv_libxml2'])
        parser.setFeature(xml.sax.handler.feature_validation, True)
        parser.setContentHandler(HelpHandler(self.textview, eob))
        #print "Parsing Help Body:\n" + "-"*80 + body  + "-"*80
        parser.parse(StringIO(body))
        
        if not eob.starts_line():
            buffer.insert(eob, "\n")
        self.scroll_to_bottom()
        
    def page_up(self):
        self.vadjustment.set_value(max(0, self.vadjustment.value -
                		       self.vadjustment.page_increment))

    def page_down(self):
        self.vadjustment.set_value(min(self.vadjustment.upper -
                		       self.vadjustment.page_size,
                		       self.vadjustment.value +
                		       self.vadjustment.page_increment))

    def insert_mathml(self, iter, mathml_xml):
        buf = self.textview.get_buffer()
        mathml_view = MathMLView()
        mathml_view.parse_string(mathml_xml)
        root = mathml_view.mathml_tree
        
        display = root.getAttribute('display', default='inline').str
        if   display == 'inline': block = False
        elif display == 'block': block = True
        else: raise ValueError("Invalid mathml display mode")

        if block:
            buf.insert(iter, "\n")
        tmpmark = buf.create_mark(None, iter, True)
        anchor = buf.create_child_anchor(iter)
        mathml_view.show()
        self.textview.add_child_at_anchor(mathml_view, anchor)
        start = buf.get_iter_at_mark(tmpmark)
        if block:
            buf.apply_tag_by_name("justify-center", start, iter)
        else: # inline
            tag = gobject.new(gtk.TextTag, rise=(-mathml_view.axis + 2)*1000)
            buf.get_tag_table().add(tag)
            buf.apply_tag(tag, start, iter)
        buf.delete_mark(tmpmark)
        if block:
            buf.insert(iter, "\n")

    def display_qa_mathml(self, question, answer):
        buffer = self.textview.get_buffer()
        eob = buffer.get_end_iter()

        buffer.insert(eob, "Query:\n")
        self.insert_mathml(eob, question)
        buffer.insert(eob, "\n")

        buffer.insert(eob, "Result:\n")
        self.insert_mathml(eob, answer)
        buffer.insert(eob, "\n")

        self.scroll_to_bottom()

    def clear(self):
        self.textview.get_buffer().set_text("")

    def print_output(self, text, msg_type):
        ## FIXME: msg_type ignored for now
        buffer = self.textview.get_buffer()
        eob = buffer.get_end_iter()
        buffer.insert_with_tags_by_name(eob, text, "message")
        if not eob.starts_line():
            buffer.insert(eob, "\n")
        self.scroll_to_bottom()
        

