#! /usr/bin/env python
import os
import subprocess
import bonobo
import ORBit, CORBA
import Bonobo
import gnumexp_conf
import atexit
import signal

ORBit.load_typelib(gnumexp_conf.KERNEL_TYPELIB)

# kernel = bonobo.get_object('OAFIID:Numexp_Kernel', 'Numexp/Kernel')

factory = None

def _start_factory():
    pipe_r, pipe_w = os.pipe()
    proc = subprocess.Popen([gnumexp_conf.KERNEL_EXE, '--private', str(pipe_w)],
                            shell=False, bufsize=1, close_fds=False)
    atexit.register(os.kill, proc.pid, signal.SIGTERM)
    os.close(pipe_w)
    ior = os.fdopen(pipe_r).readline().strip()
    return ior

def numexp_private_get():
    global factory
    if factory is None:
        ior = _start_factory()
        factory = bonobo.orb().string_to_object(ior)
    return factory.createObject('OAFIID:Numexp_Kernel')


class NumexpKernel(object):

    class Element:
        """Wrapper class for Numexp Elements, with automatic garbage collection"""
        def __init__(self, kernel, id):
            self.id     = id
            self.kernel = kernel
        def __del__(self):
            if self.id:
                self.kernel.elementUnref(self.id)
                self.id = None

    def __init__(self):
        try:
            self.kernel = bonobo.activation.activate("repo_ids.has('IDL:Numexp/Kernel:1.0')")
        except Bonobo.GeneralError:
            self.kernel = None
        if self.kernel is None:
            self.kernel = numexp_private_get()
        assert self.kernel is not None
        self._load_startup_modules()

    def get_startup_modules(self):
        try:
            conf = file(os.path.expanduser("~/.numexp-modules.conf"))
        except IOError:
            conf = file(os.path.join(gnumexp_conf.SYSCONFDIR,  "numexp-modules.conf"))
        return [module_name.rstrip() for module_name in conf if not module_name.startswith('#')]

    def set_startup_modules(self, startup_modules):
        conf = file(os.path.expanduser("~/.numexp-modules.conf"), "w")
        print >> conf, "## List of modules to load by default, one per line"
        print >> conf, "## Lines starting with # are ignored"
        for module in startup_modules:
            print >> conf, module
        conf.close()

    def _load_startup_modules(self):
        self.kernel.loadModule("stdlib") # this module is indispensable for gnumexp
        for module_name in self.get_startup_modules():
            self.kernel.loadModule(module_name.rstrip())

    def __del__(self):
        if self.kernel is not None:
            self.kernel.unref()
            self.kernel = None

    def _python_value_to_element(self, value):
        if isinstance(value, str):
            return self.kernel.parseString(value)
        elif isinstance(value, float):
            return self.kernel.elementReal(value)
        elif isinstance(value, int):
            return self.kernel.elementInt(value)

    def linear_sample(self, expression, variable, start, end, samples):
        expr = self._python_value_to_element(expression)
        var = self._python_value_to_element(variable)
        start = self._python_value_to_element(start)
        end = self._python_value_to_element(end)
        ignore_errors = self.kernel.elementInt(1);
        nsamples = self.kernel.elementInt(int(samples));
        ## Notice that the function element takes ownership of the
        ## references to the parameters.
        func     = self.kernel.elementFunction("sample", [expr, var, start, end, nsamples, ignore_errors])
        try:
            el       = self.kernel.elementEval(func)
            res      = self.kernel.elementAsAny(el)
            values   = [x.value() for x in res.value()]
        finally:
            self.kernel.elementUnref(func)
            self.kernel.elementUnref(el)
        return values

    def adaptive_sample(self, expr, var, xmin, xmax, ymin, ymax, prec):
        expr = self.kernel.parseString(expr)
        var = self.kernel.parseString(var)
        xmin = self._python_value_to_element(xmin)
        xmax = self._python_value_to_element(xmax)
        ymin = self._python_value_to_element(ymin)
        ymax = self._python_value_to_element(ymax)
        prec = self.kernel.elementInt(prec)
        ## Notice that the function element takes ownership of the
        ## references to the parameters
        func = self.Element(self.kernel,
                            self.kernel.elementFunction("adaptivesample",
                                                        [expr, var, xmin, xmax, ymin, ymax, prec]))
        el = self.Element(self.kernel, self.kernel.elementEval(func.id))
        res = self.kernel.elementAsAny(el.id)
        values = [x.value() for x in res.value()]
        return values

#     def load_script_file(self, file_name):
#         file_name_elem = self.kernel.parseString(file_name)
#         func = self.Element(self.kernel,
#                             self.kernel.elementFunction("load_file", [file_name_elem]))
#         el = self.Element(self.kernel, self.kernel.elementEval(func.id))
#         res = self.kernel.elementAsAny(el.id)
#         return res.value()
        

def test():
    kern = NumexpKernel()
    kernel = kern.kernel
    expr = kernel.parseString("x^2"); print "expr: '%s'" % (kernel.elementStr(expr),)
    var = kernel.parseString("x"); print "var: '%s'" % (kernel.elementStr(var),)
    start = kernel.elementReal(1.5); print "start: '%s'" % (kernel.elementStr(start),)
    end = kernel.elementReal(11.5); print "end: '%s'" % (kernel.elementStr(end),)
    nsamples = kernel.elementInt(100); print "nsamples: '%s'" % (kernel.elementStr(nsamples),)
    func = kernel.elementFunction("sample", [expr, var, start, end, nsamples])
    print "func: '%s'" % (kernel.elementStr(func),)
    r  = kernel.elementEval(func)
    any = kernel.elementAsAny(r)
    print "any:", any
    print "typecode:", any.typecode()
    print "value:", any.value()
    print "value2:", map(lambda x: x.value(), any.value())
    rs = kernel.elementStr(r)
    print "result =", rs


#if __name__ == '__main__':
#    test()

if __name__ == '__main__':
    print "***********", numexp_private_get(), "******************"


