#! /usr/bin/env python
from __future__ import division

import sys

if __name__ == '__main__':
    sys.path.insert(0, "..") # for debugging

import gobject
import gtk
#import gtk.gdk
import gtk.glade as glade
import gnumexp_conf
from string import zfill
import xml.dom.minidom as dom
import goocanvas
import warnings


## *sigh* I wish Python had a native enumeration class like in pygobject...
class Linemode:
    (SOLID,
     DOTTED,
     DOTDASHED,
     SHORTDASHED,
     LONGDASHED,
     DOTDOTDASHED,
     DOTDOTDOTDASHED) = range(7)

    values = (SOLID,
              DOTTED,
              DOTDASHED,
              SHORTDASHED,
              LONGDASHED,
              DOTDOTDASHED,
              DOTDOTDOTDASHED)

    to_string = {
        SOLID: 'solid',
        DOTTED: 'dotted',
        DOTDASHED: 'dotdashed',
        SHORTDASHED: 'shortdashed',
        LONGDASHED: 'longdashed',
        DOTDOTDASHED: 'dotdotdashed',
        DOTDOTDOTDASHED: 'dotdotdotdashed',
        }

    from_string = {
        'solid': SOLID,
        'dotted': DOTTED,
        'dotdashed': DOTDASHED,
        'shortdashed': SHORTDASHED,
        'longdashed': LONGDASHED,
        'dotdotdashed':        DOTDOTDASHED,
        'dotdotdotdashed': DOTDOTDOTDASHED,
        }

    to_dashes = {
        SOLID: [],
        DOTTED: [1, 2],
        DOTDASHED: [1, 4, 2],
        SHORTDASHED: [3, 2],
        LONGDASHED: [6, 2],
        DOTDOTDASHED: [1, 2, 1, 2, 4, 2],
        DOTDOTDOTDASHED: [1, 2, 1, 2, 1, 2, 4, 2],
        }

def _linemode_to_string(mode):
    return Linemode.to_string[mode]

def _linemode_from_string(mode):
    return Linemode.from_string[mode]


class Color:
    def __init__(self, *args):
        """Color(name) or Color(red, green, blue[, alpha])"""
        self.name = None
        self.red = self.green = self.blue = self.alpha = None
        if 3 <= len(args) <= 4:
            self.red = args[0]
            self.green = args[1]
            self.blue = args[2]
            if len(args) == 4:
                self.alpha = args[3]
        elif len(args) == 1:
            color = args[0]
            if color[0] == '#' and len(color) == 7:
                self.red   = int(color[1:3], 16) << 8
                self.green = int(color[3:5], 16) << 8
                self.blue  = int(color[5:7], 16) << 8
            else:
                self.name = color
        else:
            raise ValueError("Invalid number of arguments")
        if self.name is not None:
            c = gtk.gdk.color_parse(self.name)
            self.red   = c.red
            self.green = c.green
            self.blue  = c.blue


    def __str__(self):
        return '#' + zfill(hex(int(self.red >> 8))[2:], 2) + \
               zfill(hex(int(self.green >> 8))[2:], 2)     + \
               zfill(hex(int(self.blue >> 8))[2:], 2)

    def distance(self, other):

        """Calculate a number indicating how much a color differs from
        another. Range: 0-1.0"""

        return (abs(self.red   - other.red) + 
                abs(self.green - other.green) + 
                abs(self.blue  - other.blue)) / (3*65536.0)



class LineAttr:
    def __init__(self, color=Color(0,0,0), thickness=1.0,
                 dash_style=Linemode.SOLID):
        assert isinstance(color, Color)
        assert isinstance(dash_style, int)
        self.dash_style = dash_style
        self.color      = color
        self.thickness  = thickness

    def apply_to_cairo(self, cr, width_scale):
        # color
        cr.set_source_rgb(self.color.red/65535, self.color.green/65535, self.color.blue/65535)
        # dash style
        dashes = Linemode.to_dashes[self.dash_style]
        width = self.thickness*width_scale
        cr.set_dash([dash*width for dash in dashes])
        # width
        cr.set_line_width(width)

    def apply_to_goo_canvas_item(self, item, width_scale):
        assert isinstance(item, goocanvas.ItemSimple)
        if self.color.alpha is None:
            item.props.stroke_color = str(self.color)
        else:
            item.props.stroke_color_rgba = \
                (int(self.color.red/255) << 24) \
                | (int(self.color.green/255) << 16) \
                | (int(self.color.blue/255) << 8) \
                | (int(self.color.alpha/255))
        dashes = Linemode.to_dashes[self.dash_style]
        width = self.thickness*width_scale
        try:
            goocanvas.LineDash
        except AttributeError:
            warnings.warn("pygoocanvas lacks support for line dashes")
        else:
            item.props.line_dash = goocanvas.LineDash([dash*width for dash in dashes])

        item.props.line_width = width

    def get_linewidth(self, width_scale):
        return self.thickness*width_scale

    def to_xml(self, doc):
        assert isinstance(doc, dom.Document)
        elem = doc.createElement("lineattr")
        elem.setAttribute("color", str(self.color))
        elem.setAttribute("mode",  _linemode_to_string(self.dash_style))
        elem.setAttribute("width", str(self.thickness))
        return elem
    
    def from_xml(self, elem):
        assert isinstance(elem, dom.Element)
        assert elem.tagName == 'lineattr'
        self.color      = Color(elem.getAttribute("color"))
        self.dash_style = _linemode_from_string(elem.getAttribute("mode"))
        self.thickness  = float(elem.getAttribute("width"))

    def __str__(self):
        return "LineAttr(thickness=%f, dash_style=%s, color='%s')" % \
               (self.thickness, _linemode_to_string(self.dash_style),
                str(self.color))

    def distance(self, other):
        """Calculate a number indicating how much a lineattr differs from another"""
        return abs(self.dash_style - other.dash_style) + \
               abs(self.thickness - other.thickness) + \
               10*Color.distance(self.color, other.color)
        

class LineAttrPreview(gtk.DrawingArea):

    __gtype_name__ = 'LineAttrPreview'
    
    def __init__(self, attr, border=1, width=-1, height=-1, background=None):
        assert isinstance(attr, LineAttr)
        assert background is None or isinstance(background, gtk.gdk.Color)
        gtk.DrawingArea.__init__(self)
        self.attr = attr
        self.draw_border = border
        self.background = background
        self.set_size_request(width, height)

    def do_expose_event(self, event):
        cr = self.window.cairo_create()
        self.attr.apply_to_cairo(cr, self.allocation.height/100.0)
        cr.move_to(self.allocation.width*0.1, self.allocation.height*0.5)
        cr.line_to(self.allocation.width*0.9, self.allocation.height*0.5)
        cr.stroke()
        ## TODO: border

class LineAttrCellRenderer(gtk.CellRenderer):
    __gtype_name__ = "LineAttrCellRenderer"

    __gproperties__ = { 'lineattr' :
                        (object,
                         'LineAttr',
                         'LineAttr (line attributes)',
                         gobject.PARAM_WRITABLE)
                        }
    def __init__(self):
        gtk.CellRenderer.__init__(self)
        self.attr = None

    def do_get_size(self, widget, cell_area):
        x_offset = 0
        y_offset = 0
        width = 100
        height = 10
        return x_offset, y_offset, width, height

    def do_render(self, window, widget, background_area, cell_area, expose_area, flags):
        cr = window.cairo_create()
        cr.translate(cell_area.x, cell_area.y)
        cr.move_to(cell_area.width*0.1, cell_area.height*0.5)
        cr.line_to(cell_area.width*0.9, cell_area.height*0.5)
        self.attr.apply_to_cairo(cr, cell_area.height/100)
        cr.stroke()

    def do_set_property(self, pspec, value):
        if pspec.name == 'lineattr':
            assert isinstance(value, LineAttr)
            self.attr = value
        else:
            raise TypeError


class LineAttrSelector(gtk.Expander):
    __gsignals__ = {
        'lineattr-set': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE,
                         ())
    }
    def __init__(self, attr, label=None, use_underline=False):
        gtk.Expander.__init__(self)
        self.attr = attr
        self.set_spacing(4)

        self.xml = glade.XML(gnumexp_conf.GLADEDIR + "/lineattr.glade2",
                             "lineattr_controls")

        self.spin = self.xml.get_widget("line_width")
        self.xml.get_widget("width label").set_mnemonic_widget(self.spin)

        self.color_picker = self.xml.get_widget("line_color")
        self.xml.get_widget("color label").set_mnemonic_widget(self.color_picker)

        self.combobox = self.xml.get_widget("line_style")
        self.xml.get_widget("line style label").set_mnemonic_widget(self.combobox)
        listmodel = gtk.ListStore(object)
        for dash_style in Linemode.values:
            iter_ = listmodel.append()
            attr = LineAttr(thickness=20.0, dash_style=dash_style, color=Color(0,0,0))
            listmodel.set(iter_, 0, attr)
        cell = LineAttrCellRenderer()
        self.combobox.pack_start(cell, True)
        self.combobox.add_attribute(cell, 'lineattr', 0)
        self.combobox.set_model(listmodel)

        ## Expander label
        alignment = gtk.Alignment(0.5, 0.5, 0.75, 0.25)
        alignment.show()
        alignment.set_border_width(1)
        self.preview = LineAttrPreview(self.attr, border=0, width=120, height=16)
        self.preview.show()
        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        alignment.add(frame)
        frame.show()
        frame.add(self.preview)
        hbox = gtk.HBox(False, 4)
        hbox.show()
        if label is not None:
            lb = gtk.Label(label)
            lb.show()
            lb.set_use_underline(use_underline)
            hbox.pack_start(lb)
        hbox.pack_start(alignment)
        self.set_label_widget(hbox)

        ## Expander content
        hbox = gtk.HBox(False, 4)
        hbox.show()
        lb = gtk.Label(' '*8)
        lb.show()
        hbox.pack_start(lb, False, False)
        hbox.add(self.xml.get_widget("lineattr_controls"))
        self.add(hbox)

        self.xml.signal_autoconnect({
            'lineattr_window_preview_update': self.__changed
            })

        ## set initial values in widgets
        self.set_attr(self.attr)


    def __set_dash_style(self, opmenu):
        item = opmenu.get_menu().get_active()
        style = item.get_data("dash-style")
        self.attr.dash_style = style
        self.__preview_update()

    def __changed(self, *args):
        if self.__block_changes:
            return
        color = self.color_picker.get_color()
        self.attr.color = Color(color.red, color.green, color.blue)
        self.attr.thickness = self.spin.get_value()
        attr, = self.combobox.get_model().get(self.combobox.get_active_iter(), 0)
        self.attr.dash_style = attr.dash_style
        self.emit("lineattr-set")

    def do_lineattr_set(self):
        if self.preview.window is not None:
            self.preview.queue_draw()

    def set_attr(self, attr):
        ## block change signals from widgets
        self.__block_changes = True

        self.attr = attr
        self.preview.attr = self.attr

        listmodel = self.combobox.get_model()
        i = listmodel.get_iter_root()
        while True:
            tmp_attr, = listmodel.get(i, 0)
            if tmp_attr.dash_style == self.attr.dash_style:
                self.combobox.set_active_iter(i)
                break
            i = listmodel.iter_next(i)

        self.color_picker.set_color(gtk.gdk.Color(self.attr.color.red,
                                                  self.attr.color.green,
                                                  self.attr.color.blue))
        self.spin.set_value(self.attr.thickness)
        self.__block_changes = False

        self.emit("lineattr-set")
        #if self.preview.window is not None:
        #    self.preview.update()
gobject.type_register(LineAttrSelector)
        


def _test_selector():
    attr = LineAttr(thickness=10.0, dash_style=Linemode.LONGDASHED, color=Color(0,0,65535))
    win = gtk.Window()
    #bt = LineAttrButton(attr)
    bt = LineAttrSelector(attr, "Line _Style:", True)
    win.add(bt)
    win.set_border_width(12)
    bt.show()
    win.connect('destroy', lambda w: gtk.main_quit())
    win.show()
    gtk.main()


def _test_xml():
    doc  = dom.Document()
    attr = LineAttr(thickness=10.0, dash_style=Linemode.LONGDASHED, color=Color(0,0,65535))
    elem = attr.to_xml(doc)
    print elem
    print elem.toxml()
    attr1 = LineAttr()
    attr1.from_xml(elem)
    print str(attr)
    print str(attr1)
    doc.appendChild(elem)
    doc.writexml(file("/tmp/lineattr.xml", "w"))

def _test_lineattr_preview():
    w = gtk.Window()
    attr = LineAttr(thickness=10.0, dash_style=Linemode.LONGDASHED, color=Color(0,0,65535))
    preview = LineAttrPreview(attr)
    w.add(preview)
    w.show_all()
    gtk.main()

if __name__ == '__main__':
    gnumexp_conf.GLADEDIR = "."

    _test_selector()
    #_test_xml()
    #_test_lineattr_preview()

