import warnings
import itertools
from misc import *
#from kernel import *
import plotobject_registry
from dialogs.plotobject import PlotObjectDialog
import dialogs.util
import xml.dom.minidom as dom
import types, traceback
import ORBit, CORBA
from gettext import gettext as _
def N_(message): return message
import goocanvas


class FunctionObjectDialog(PlotObjectDialog):
    def __init__(self, kernel, editmode):
        PlotObjectDialog.__init__(self, "function plot object dialog", editmode)
        #self.name = "Function"

        self.expression_wid = dialogs.util.ExpressionEntry(kernel)
        self.expression_wid.show()
        self.xml.get_widget("expression container").add(self.expression_wid)
        self.xml.get_widget("expression label").set_mnemonic_widget(self.expression_wid)

        self.independent_var_wid = dialogs.util.VariableEntry(self.expression_wid)
        self.independent_var_wid.show()
        self.xml.get_widget("variable container").add(self.independent_var_wid)
        self.xml.get_widget("variable label").set_mnemonic_widget(self.independent_var_wid)

        ## set default legend text
        self.xml.get_widget("func caption").set_text("%e")

    def get_values(self):
        rv = PlotObjectDialog.get_values(self)
        rv.update({
            'type': "Function",
            'contents': {
                'expression':                 self.expression_wid.get_text(),
                'independent-variable': self.independent_var_wid.get_value(),
                }
            })
        return rv

    def set_values(self, values):
        PlotObjectDialog.set_values(self, values)
        values = values['contents']
        self.expression_wid.set_text(values['expression'])
        self.independent_var_wid.set_value(values['independent-variable'])

    def validate(self):
        '''Make a quick validation (just see if there are no missing fields)
        If not valid, give focus to the invalid widget and return False,
        else return True'''

        if not self.expression_wid.validate():
            self.expression_wid.grab_focus()
            return False

        if not self.independent_var_wid.get_value():
            ## http://bugzilla.gnome.org/show_bug.cgi?id=171912
            #self.independent_var_wid.grab_focus()
            self.independent_var_wid.mnemonic_activate(True)
            return False
        
        return True

class Function(PlotObject):

    dialog_class = FunctionObjectDialog

    def __init__(self, values, plot_area):
        self.expression = None
        self.variable = None
        self.tree = None
        self.bbox = None
        self.samples = None
        self.mathml = '<math><mtext></mtext></math>'
        self.canvas_item = None
        self.caption = "%e"
        super(Function, self).__init__(values, plot_area)
        
    def get_canvas_item(self):
        '''Return the canvas item that renderes this plot object'''
        if self.canvas_item is None:
            self.inner_path = goocanvas.Path()
            self.outer_path = goocanvas.Path(pointer_events=(goocanvas.EVENTS_STROKE_MASK
                                                             |goocanvas.EVENTS_FILL_MASK
                                                             |goocanvas.EVENTS_PAINTED_MASK))
            self.canvas_item = goocanvas.Group()
            self.canvas_item.add_child(self.outer_path)
            self.canvas_item.add_child(self.inner_path)
        return self.canvas_item

    def get_values(self):
        values = PlotObject.get_values(self)
        values['contents'] = { 'expression': self.expression,
                	       'independent-variable': 	self.variable,
                	       }
        values['type'] = 'Function'
        return values

    def set_values(self, values):
        PlotObject.set_values(self, values)
        values = values['contents']
        self.expression = values['expression']
        self.variable = values['independent-variable']
        self.update_function()
        self.__update_mathml()
        self.draw()

    def __update_mathml(self):
        try:
            el = self.kernel.kernel.parseString(self.expression)
            expression_mathml = self.kernel.kernel.elementToMathML(el)
            self.kernel.kernel.elementUnref(el)
        except Exception, ex:
            expression_mathml = '<merror>error converting expression to MathML: %s</merror>' % str(ex)
        expression_mathml = expression_mathml.decode("utf-8")
        #print "MathML: ", `expression_mathml`
        variable_mathml = '<mi>' + self.variable + '</mi>'
        self.mathml = self._label_to_mathml(self.caption,
                			    {'%e': expression_mathml,
                			     '%v': variable_mathml})
    
    def __del__(self):
        if self.tree:
            self.kernel.kernel.elementUnref(self.tree)
        self.tree = None

    def update_samples(self):
        try:
            self.sample()
        except CORBA.UserException, ex:
            self.samples = None
            err = str(ex.__class__) + ": " + ex.detail
            dialog = gtk.MessageDialog(
                flags          = gtk.DIALOG_DESTROY_WITH_PARENT,
                type           = gtk.MESSAGE_ERROR,
                buttons        = gtk.BUTTONS_OK,
                message_format = err)
            dialog.set_property("has-separator", gtk.FALSE)
            self.plot_area.view.set_transient_for(dialog)
            dialog.run()
            dialog.destroy()
            return
        except Exception, ex:
            self.samples = None
            dialog = gtk.MessageDialog(
                flags          = gtk.DIALOG_DESTROY_WITH_PARENT,
                type           = gtk.MESSAGE_ERROR,
                buttons        = gtk.BUTTONS_OK,
                message_format = _("Unknown error while evaluating expression:\n") +
                                 str(ex.__class__))
            self.plot_area.view.set_transient_for(dialog)
            dialog.run()
            dialog.destroy()
            raise
        #self.queue_draw()
        #self.update_control_points()

    def math_range_changed(self):
        '''Called when the math coordinates range has changed in the plot area'''
        self.update_samples()
        self.draw()

    def screen_range_changed(self):
        '''Called when the screen coordinates range has changed in the plot area'''
        self.draw()

    def update_function(self):
        self.update_samples()

    def sample(self):
        return self._adaptive_sample()

    def _linear_sample(self):
        if self.expression is None:
            self.samples = None
            return
        nsamples = 200
        range_ = self.plot_area.range
        yvalues = self.kernel.sample(self.expression, self.variable,
                                     range_.x1, range_.x2, nsamples)
        
        xvalues = self.kernel.sample(self.variable, self.variable,
                                     range_.x1, range_.x2, nsamples)
        
        # Combine X and Y values and store the result
        self.samples = zip(xvalues, yvalues)

    def _adaptive_sample(self):
        if self.expression is None:
            self.samples = None
            return

        range_ = self.plot_area.range
        values = self.kernel.adaptive_sample(self.expression, self.variable,
                                             range_.x1, range_.x2,
                                             range_.y1, range_.y2,
                                             3000)
        self.samples = list(itertools.izip(
            itertools.islice(values, 0, len(values), 2),
            itertools.islice(values, 1, len(values), 2)))


    def draw(self):
        if self.samples is None or self.state is None:
            return
        self.get_canvas_item()

        x1 = None
        y1 = None
        path = []
        for (x, y) in self.samples:
            if y is None: # discontinuity?
                x1 = None
                x2 = None
                continue
            if x1 is None or y1 is None:
                x1, y1 = self.plot_area.range2bbox.transform_point(x, y)
                path.append("M %f %f" % (x1, y1))
            else:
                x2, y2 = self.plot_area.range2bbox.transform_point(x, y)
                path.append("L %f %f" % (x2, y2))
                x1, y1 = x2, y2
        path_data = ' '.join(path)
        
        for lineattr, canvas_path in ((self.outer_lineattr, self.outer_path),
                                      (self.inner_lineattr, self.inner_path)):
            if lineattr.color.alpha == 0:
                canvas_path.props.visibility = goocanvas.ITEM_INVISIBLE
            else:
                canvas_path.props.visibility = goocanvas.ITEM_VISIBLE
            lineattr.apply_to_goo_canvas_item(canvas_path, self.plot_area.width_scale)
            canvas_path.props.data = path_data


    def get_legend_sample(self):
        sample = goocanvas.Path()
        ## FIXME: the 0.1 pixel offset in y2 is to make sure the line
        ## is not perfectly horizontal, thus working aroung bug
        ## https://bugs.freedesktop.org/show_bug.cgi?id=7965
        ## Should be removed once cairo gets fixed or a goocanvas
        ## workaround is added.
        sample.props.data = "M 0 0 L 30 0.1"
        self.lineattr.apply_to_goo_canvas_item(sample, self.plot_area.width_scale)
        return sample
    
    def to_xml(self, doc):
        assert isinstance(doc, dom.Document)
        elem = PlotObject.to_xml(self, doc)
        elem.setAttribute("type", "Function")
        expr = doc.createElement("expression")
        expr.appendChild(doc.createTextNode(self.expression))
        elem.appendChild(expr)
        var = doc.createElement("variable")
        var.setAttribute("name", self.variable)
        elem.appendChild(var)
        return elem

    def from_xml(self, elem, plotarea):
        PlotObject.from_xml(self, elem, plotarea)
        assert elem.getAttribute("type") == 'Function'
        expr = elem.getElementsByTagName('expression')[0]
        self.expression = ""
        for node in expr.childNodes:
            if node.nodeType == node.TEXT_NODE:
                self.expression += node.data
        self.expression = self.expression.strip().encode()
        var = elem.getElementsByTagName('variable')[0]
        self.variable = var.getAttribute("name").encode()
        self.__update_mathml()

    def get_y_range(self):
        if self.samples is None: return None
        yvalues = [p[1] for p in self.samples]
        return min(yvalues), max(yvalues)
        

plotobject_registry.type_name_to_class_mapping[N_('Function')] = Function
__all__ = ['Function']
#print "'Function' plot object has been registered"
