#! /usr/bin/env python
#from warnings import warn
from __future__ import division
import cairo
import goocanvas
import gtk.gdk
import gobject
from itertools import izip
from goomathml import GooMathML, GooMathMLView


LEGEND_BORDER_SIZE = 8
LEGEND_SPACING = 6



class LegendItem(goocanvas.Group):

    def __init__(self, plot_object, legend):
        goocanvas.Group.__init__(self)
        self.legend = legend
        self.plot_object = plot_object
        self.sample = plot_object.get_legend_sample()
        self.add_child(self.sample)
        self.label = GooMathML(x=50, y=0, font_size=12)
        self.add_child(self.label)
        legend.add_child(self)
        self.label.props.pointer_events = goocanvas.EVENTS_NONE
        self.sample.props.pointer_events = goocanvas.EVENTS_NONE
        self.update()

        label_view = legend.plot_area.view.canvas_view.get_item_view(self.label)
        label_bounds = label_view.get_bounds()
        label_height = label_bounds.y2 - label_bounds.y1
        self.sample.translate(0, label_height/2)
        self.view = legend.plot_area.view.canvas_view.get_item_view(self)

    def update(self):
        self.label.props.mathml = self.plot_object.mathml

class Legend(goocanvas.Group):
    class DragState:
        ## legend position before drag start
        x0 = None
        y0 = None

        ## pointer position when drag starts
        pointer_x0 = None
        pointer_y0 = None
    
    def __init__(self, plot_area):
        goocanvas.Group.__init__(self)
        plot_area.add_child(self)
        self.plot_area = plot_area
        self.items = {}

        ## position in pixels
        self.x = 100
        self.y = 50

        ## relative position
        self.rel_x = None
        self.rel_y = None
        self.set_position(0.1, 0.1)
        
        self.drag = None # or DragState

        gdkcolor = gtk.gdk.color_parse("ivory")
        alpha = 0xC000
        fill_color_rgba = ((int(gdkcolor.red/256) << 24)
                           | (int(gdkcolor.green/256) << 16)
                           | (int(gdkcolor.blue/256) << 8)
                           | (int(alpha/256)))
        
        self.rectangle = goocanvas.Rect(x=0, y=0,
                                        width=100, height=100,
                                        line_width=2.0,
                                        radius_x=4.0,
                                        radius_y=4.0,
                                        stroke_color="black",
                                        fill_color_rgba=fill_color_rgba)
        self.add_child(self.rectangle)

        self.plot_area.connect("plot-object-added", self._plot_object_added)
        self.plot_area.connect("plot-object-removed", self._plot_object_removed)
        self.plot_area.connect("plot-object-modified", self._plot_object_modified)

        rect_view = plot_area.view.canvas_view.get_item_view(self.rectangle)
        rect_view.connect("button-press-event", self.on_rect_button_press)
        rect_view.connect("enter-notify-event", self.on_rect_enter)
        rect_view.connect("leave-notify-event", self.on_rect_leave)
        rect_view.connect("button-release-event", self.on_rect_button_release)
        rect_view.connect("motion-notify-event", self.on_rect_motion_notify)

        self._layout()

    def set_position(self, rel_x, rel_y):
        self.rel_x = rel_x
        self.rel_y = rel_y
        self.update_position()

    def update_position(self):
        bbox = self.plot_area.bbox
        x = self.rel_x*(bbox.x2 - bbox.x1) + bbox.x1
        y = self.rel_y*(bbox.y2 - bbox.y1) + bbox.y1
        self._move_to(x, y)

    def _update_relative_position(self):
        bbox = self.plot_area.bbox
        self.rel_x = (self.x - bbox.x1) / (bbox.x2 - bbox.x1)
        self.rel_y = (self.y - bbox.y1) / (bbox.y2 - bbox.y1)

    def _move_to(self, x, y):
        self.x = x
        self.y = y
        M = cairo.Matrix()
        M.translate(self.x, self.y)
        self.set_transform(M)

    def _plot_object_added(self, plot_area, plot_object):
        item = LegendItem(plot_object, self)
        self.items[plot_object] = item
        self._layout()

    def _plot_object_removed(self, plot_area, plot_object):
        item = self.items[plot_object]
        del self.items[plot_object]
        self.remove_child(item)
        self._layout()

    def _plot_object_modified(self, plot_area, plot_object):
        item = self.items[plot_object]
        item.update()
        self._layout()

    def _layout(self):
        if not self.items:
            self.rectangle.props.visibility = goocanvas.ITEM_INVISIBLE
            return
        else:
            self.rectangle.props.visibility = goocanvas.ITEM_VISIBLE
        widths = []
        heights = []
        for item in self.items.itervalues():
            bounds = item.view.get_bounds()
            widths.append(bounds.x2 - bounds.x1)
            heights.append(bounds.y2 - bounds.y1)
        y = LEGEND_BORDER_SIZE
        for item, height in izip(self.items.itervalues(), heights):
            M = cairo.Matrix()
            M.translate(LEGEND_BORDER_SIZE, y)
            item.set_transform(M)
            y += height + LEGEND_SPACING
        self.rectangle.props.width = max(widths) + 2*LEGEND_BORDER_SIZE
        self.rectangle.props.height = y + LEGEND_BORDER_SIZE - LEGEND_SPACING
        
    def on_rect_enter(self, rect_view, target, event):
        item = target.get_item()
        item.props.line_width = 4.0
        return False

    def on_rect_leave(self, rect_view, target, event):
        item = target.get_item()
        item.props.line_width = 2.0
        return False

    def on_rect_button_press(self, rect_view, target, event):
        if self.drag is not None:
            return False
        self.drag = self.DragState()
        self.drag.x0 = self.x
        self.drag.y0 = self.y

        canvas = rect_view.get_canvas_view()
        canvas.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))

        ## workaround a goocanvas bug
        x, y = canvas.convert_from_item_space(rect_view, event.x, event.y)

        self.drag.pointer_x0 = x
        self.drag.pointer_y0 = y
        return False

    def on_rect_button_release(self, rect_view, target, event):
        if self.drag is None:
            return False
        self.drag = None
        canvas = rect_view.get_canvas_view()
        canvas.window.set_cursor(None)
        self._update_relative_position()
        return False

    def on_rect_motion_notify(self, rect_view, target, event):
        if self.drag is None:
            return False

        ## workaround a goocanvas bug
        x, y = rect_view.get_canvas_view().convert_from_item_space(rect_view, event.x, event.y)

        self._move_to(self.drag.x0 + (x - self.drag.pointer_x0),
                      self.drag.y0 + (y - self.drag.pointer_y0))

